/*
 * This program is written to use Jacobi iteration to solve Laplace equations.
 * It is set up to run on 4 processors in a 2 by 2 grid.  The region to
 * be solved is a 16 by 16 array of grid points, including the boundary
 * values.  So each processor will have 8 by 8 of the grid points,
 * centered in a 10 by 10 array, where the outer ring of points is the
 * "guard ring" that allows storage of neighbor points from other
 * processors.  This array is called phi in each processor; there is
 * also an array oldphi to hold old values during each iteration step,
 * and an array called mask, which is true on the boundary values and
 * false elsewhere.  
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define xprocessors 3
#define yprocessors 3

#define xpoints 4
#define ypoints 4

#define NDIM 2

#define no_ants 100

struct ants
{ int ant_no;
  int no_of_moves;
};

typedef struct ants ANT;
typedef ANT *LINK;


main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, leftneighbor, rightneighbor;
    int bottomneighbor, topneighbor;
    int coords[NDIM], src;
    int rankx, ranky;
    
    /* the ant is identified by the ant number, and each ant
       has the number of moves it made, initially zero. For each move,
       the number of moves is incremented by one
    */
    
    LINK new_ant[no_ants];
    int rand_no;
    int gen_no;

    int block[ypoints][xpoints]; /* blocks in the grid */
    int value, size, i, j, m, n, reorder;
    int dims[2], periods[2];
    
    int done = 0;     /* test flag */
    int iter = 0;     /* iteration counter */
    
    MPI_Comm comm2d;
    MPI_Datatype  contig, strided;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    /* .... set up communication ....*/

    periods[0] = 0;
    periods[1] = 0;

    reorder = 0;

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		    &comm2d);
    MPI_Cart_coords(comm2d, rank, 2, coords);

    ranky = coords[0];
    rankx = coords[1];

    MPI_Cart_shift(comm2d, 0, 1, &bottomneighbor, &topneighbor);
    MPI_Cart_shift(comm2d, 1, 1, &leftneighbor, &rightneighbor);

    /* .... set up the grid ....*/

/*   0 = nest
     1 = food
     2 = ant
     3 = chemical
     4 = empty
*/
      /* Initialize all blocks, assign the nest and the food */
      for (j = 0; j < xpoints; j++)
      { for (i = 0; i < ypoints; i++)
	{ block[i][j] = 4;
        }
      }

      block[1][1] = 0; /* nest */
      block[0][1] = 1; /* food */
      block[2][2] = 1; /* food */

    /* process 0, creates all ants and start their movement.
       the other processors, receive the ant, checks if the is food.
       If the food is found it picks the food returns to the nest, 
       Otherwise, the processor determines the next route, and send the ant there.
     
     */
      srand(time(NULL));
      
    if (rankx == 0) 
    { m = 1; n = 1; /* positions of nest */
      for (i = 1; i < no_ants; i++)
      { new_ant->ant_no = i; 
        new_ant->no_of_moves = 0;
        rand(rand_no);
        gen_no = (random()%4);
	
        if (gen_no == 0)
        {  MPI_Sendrcv(&new_ant, 1, LINK, topneighbor, 1, new_ant, 1,
             LINK, bottomneigbor, 1, comm2d,&status);
        }
        if (gen_no == 1)
        {  MPI_Sendrcv(&new_ant, 1, LINK, rightneighbor, 1, new_ant, 1,
             LINK, leftneigbour, 1, comm2d,&status); 
        }
        if (gen_no == 2)
        {  MPI_Sendrcv(&new_ant, 1, LINK, bottomneighbor, 1, new_ant, 1,
             LINK, topneighbor, 1, comm2d,&status); 
        }
        if (gen_no == 3)
        {  MPI_Sendrcv(&new_ant, 1, LINK, leftneighbor, 1, new_ant, 1,
             LINK, rightneighbor, 1, comm2d,&status); 
        }

      }
    }
    else
    {  new_ant->no_of_moves += 1;

       if (new_ant->no_of_moves > 20)
       { /* ant returns to the nest */

       }
/*       if ( food found)
       { take it to nest using shortest route 
       }
*/

/*     if (no food, but chemical can be smelled)
       { follow the direction of the chemical
       }
       /* generate the next move */
       rand(rand_no);
       gen_no = rand_no%4;
       if (gen_no == 1)
        {  MPI_Sendrcv(&new_ant, 1, LINK, rightneighbor, 1, new_ant, 1,
             LINK leftneigbour, 1, comm2d,&status); 
        }
        if (gen_no == 2)
        {  MPI_Sendrcv(&new_ant, 1, LINK, bottomneighbor, 1, new_ant, 1,
             LINK topneighbor, 1, comm2d,&status); 
        }
        if (gen_no == 3)
        {  MPI_Sendrcv(&new_ant, 1, LINK, leftneighbor, 1, new_ant, 1,
             LINK rightneighbor, 1, comm2d,&status); 
        }
    }
    

    MPI_Finalize();
}
